#include "smtp.h"
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

/* smtp->jabber */

/* ensure that data is 7 bit */
char *smtp_mime_check7bit(smtps s, mimectx ctx, char *str) {
    char *ptr;

    if (!(ctx->mimeEncoding == MIME_7BIT || ctx->mimeEncoding == MIME_QUOTED || ctx->mimeEncoding == MIME_BASE64))
	return str;

    for (ptr=str; ptr && *ptr; ptr++)
	*ptr |= 0x7f;

    return str;
}

/* convert a string to UTF-8 encoding */
char *smtp_mime_toUTF8(smtps s, mimectx ctx, char *str) {
    char *ptr;

    /* XXX this is not correct handling */
    for (ptr=str; ptr && *ptr; ptr++)
	if (*ptr>126 || *ptr<1)
	    *ptr = ' ';

    return str;
}

/* decode a quoted printable string line */
char *smtp_mime_decode_quoted(smtps s, mimectx ctx, char *str) {
    return smtp_mime_toUTF8(s, ctx, str);
}

/* decode a base64 encoded string */
char *smtp_mime_decode_base64(smtps s, mimectx ctx, char *str) {
    return smtp_mime_toUTF8(s, ctx, str);
}

/* decode a mime encoded string */
char *smtp_mime_decode(smtps s, mimectx ctx, char *str) {
    if (!ctx->mimeUsed)
	return str;

    switch (ctx->mimeEncoding) {
	case MIME_7BIT:
	    smtp_mime_check7bit(s, ctx, str);
	    return str;
	case MIME_8BIT:
	    return smtp_mime_toUTF8(s, ctx, str);
	case MIME_BINARY:
	    /* XXX how to handle binary encoded data? */
	    return smtp_mime_toUTF8(s, ctx, str);
	case MIME_QUOTED:
	    return smtp_mime_decode_quoted(s, ctx, str);
	case MIME_BASE64:
	    return smtp_mime_decode_base64(s, ctx, str);
	default:
	    log_debug(ZONE, "invalid mimeEncoding in mimectx: %i", ctx->mimeEncoding);
    }
    return str;
}

/* send an outgoing line to the connection */
void smtp_status(smtps s, int status, char *data)
{
    char istr[10] = "";
    char *out;
    pool p;
    spool line;

    sprintf(istr,"%d ",status);

    p = pool_new();
    line = spool_new(p);
    spooler(line,istr,data,"\r\n",line);

    out = spool_print(line);
    log_debug("SMTP-OUT",out);
    mio_write(s->m,NULL,out,-1);
}

/* decode the incoming commands */
void smtp_in_cmd(smtps s, int cmd, char *addr)
{
    jid id;
    char *bkt;

    /* we only accept 7 bit characters in SMTP commands */
    for(bkt = addr; bkt && *bkt; bkt++)
	if(*bkt > 126 || *bkt < 32)
	    *bkt = ' ';

    if(addr != NULL)
    {
        bkt = strstr(addr,"<");
        if(bkt != NULL) addr = bkt + 1;
        bkt = strstr(addr,">");
        if(bkt != NULL) *bkt = '\0';
    }

    log_debug(ZONE,"command %d with data %s",cmd,addr);

    switch(cmd)
    {
        case CMD_HELO:
            smtp_status(s, 250, "Hi.");
            break;
        case CMD_MAIL:
            s->from = pstrdup(s->p,addr);
            smtp_status(s, 250, "Ok");
            break;
        case CMD_RCPT:
            id = smtp_mx2(s->i,jid_new(s->p,addr));
            if(id != NULL)
            {
                if(s->rcpt == NULL)
                    s->rcpt = id;
                else
                    s->rcpt = jid_append(s->rcpt,id);
                smtp_status(s, 250, "Ok");
            }else{
                smtp_status(s, 550, "Bad Address");
            }
            break;
        case CMD_DATA:
            s->phase = PHASE_DATA;
            s->x = xmlnode_new_tag_pool(s->p,"mail");
            smtp_status(s, 354, "Feed me");
            break;
        case CMD_RSET:
            s->x = NULL;
            s->rcpt = NULL;
            smtp_status(s, 250, "Boring");
            break;
        case CMD_NOOP:
            smtp_status(s, 250, "Boring");
            break;
        case CMD_QUIT:
            smtp_status(s, 221, "Bye");
            break;
	default:
            smtp_status(s, 500, "Duh");
    }
}

int smtp_cmd_parse(char *cmd)
{
    log_debug(ZONE, "parsing command %s", cmd);

    if(strcasecmp(cmd,"HELO") == 0) return CMD_HELO;
    if(strcasecmp(cmd,"MAIL") == 0) return CMD_MAIL;
    if(strcasecmp(cmd,"RCPT") == 0) return CMD_RCPT;
    if(strcasecmp(cmd,"DATA") == 0) return CMD_DATA;
    if(strcasecmp(cmd,"NOOP") == 0) return CMD_NOOP;
    if(strcasecmp(cmd,"RSET") == 0) return CMD_RSET;
    if(strcasecmp(cmd,"QUIT") == 0) return CMD_QUIT;

    return CMD_NONE;
}

char *smtp_in_getaddr(pool p, char *str)
{
    char *cur, *top;
    char invalid[] = "\"'<> \t,";
    int isaddr = 0;

    log_debug(ZONE,"searching for an address in: %s",str);

    for(top = cur = pstrdup(p,str);cur != NULL && *cur != '\0';cur++)
    {
        /* can't grok invalid chars in an address */
        if(index(invalid,*cur) != NULL)
        {
            /* we've hit an invalid char at the end of a valid address, snip it and quit */
            if(isaddr)
            {
                *cur = '\0';
                break;
            }
            top = cur + 1;
            continue;
        }

        /* yay, a @ after some valid chars, it's an address! (or so we hope) */
        if(*cur == '@')
        {
            *cur = '%'; /* escape it, of course */
            isaddr = 1;
            continue;
        }
    }

    if(isaddr)
    {
        log_debug(ZONE,"we've found something we think is an address: %s",top);
        return top;
    }

    log_debug(ZONE,"nothing of concequence found");
    return NULL;
}

/* when the connection dies, clean up the mess */
void smtp_in_cleanup(void *arg)
{
    smtps s = (smtps)(arg);
 
    log_debug(ZONE,"smtp_in_cleanup");
    if(s != NULL)
    {
        free(s->buffer);
    }
}

/* extract a header line from the received message */
char *smtp_in_header(smtps s, mimectx ctx, const char *msg, const char *header, int structured)
{
    const char *linebegin = msg;
    const char *lineend = strchr(linebegin, '\n');
    char *tbuf;
    size_t hlen;
    char *ptr;

    if (header == NULL)
	return NULL;

    hlen = strlen(header);

    /* XXX We don't handle headerlines that are splitted to multible lines */
    while (linebegin && lineend && lineend>linebegin)
    { /* there is another headerline to process */

	/* is it the header we are looking for? */
	if (strncasecmp(linebegin, header, hlen)==0 && linebegin[hlen]==':')
	{ /* we found the header */
	    linebegin = linebegin+hlen+1;
	    /* skip space at the beginning of header content */
	    while (*linebegin == ' ') linebegin++;
	    /* skip space at the end of the header content */
	    while (lineend[-1] == ' ') lineend--;

	    /* empty header? */
	    if (lineend <= linebegin)
		return pstrdup(s->p, "");

	    /* return header content */
	    hlen = lineend - linebegin + 1;
	    tbuf = pmalloc(s->p, hlen+1);
	    tbuf[hlen] = '\0';
	    strncpy(tbuf, linebegin, hlen);

	    /* headers are not allowed to contain non-ASCII in encoded form */
	    for (ptr = tbuf; *ptr; ptr++)
		if (*ptr > 126 || *ptr < 32) *ptr = ' ';

	    /* content in brackets is comment if it is a structured header */
	    while (structured && (ptr = strchr(tbuf, '(')) != NULL) {
		char *commentend = strchr(ptr, ')');
		if (commentend == NULL)
		    *ptr='\0';
		else {
		    size_t copylen;
		    size_t i;
		    copylen = strlen(commentend);

		    for (i=0; i<copylen; i++)
			ptr[i] = commentend[i+1];
		}
	    }

	    /* XXX unescape MIME here */

	    return tbuf;
	}

	/* continue with next header line */
	linebegin = lineend + 1;
	lineend = strchr(linebegin, '\n');
    }

    /* no header found */
    return NULL;
}

void smtp_in_message(smtps s)
{
    jid cur;
    xmlnode x;
    char *from;
    char *msg;
    char *body;
    char *ptr;
    _mimectx ctx;

    /* init the mime context to not using MIME */
    ctx.mimeUsed = 0;
    ctx.mimeEncoding = MIME_7BIT;
    ctx.mimeCharset = "us-ascii";

    log_debug(ZONE,"yay, I got a message! [%s]",xmlnode2str(s->x));

    msg = spool_print(s->message);

    /* call the spam detection */
    if (s->i->spamcheck) {
	pid_t checkpid;
	int pipefd[2];

	if (pipe(pipefd)) {
	    log_error(ZONE, "could not create pipe to feed spamcheck script");
	    smtp_status(s,430,"problem executing spam control");
	    return;
	}

	log_debug(ZONE, "calling %s for spam check", s->i->spamcheck);

	checkpid = fork();

	if (checkpid == -1) {
	    close(pipefd[0]);
	    close(pipefd[1]);
	    log_error(ZONE, "cannot fork child process for spam checking");
	    smtp_status(s,430,"problem executing spam control");
	    return;
	} else if (checkpid == 0) {
	    /* inside child process */

	    /* close writing end of pipe */
	    close(pipefd[1]);

	    /* dup pipe to standard input */
	    dup2(pipefd[0], 0);
	    close(pipefd[0]);

	    /* execute the script */
	    execlp(s->i->spamcheck, NULL);

	    /* execlp only returns on error */
	    _exit(42);
	} else {
	    /* inside parent process */
	    int status;
	    ssize_t written = 0;
	    ssize_t wstat;
	    size_t towrite;

	    /* close reading end of the pipe */
	    close(pipefd[0]);

	    /* feed message to the pipe */
	    towrite = strlen(msg);
	    while (towrite > 0) {
		wstat = write(pipefd[1], msg+written, towrite-written);
		if (wstat == -1) {
		    log_error(ZONE, "problem writing to spam check pipe");
		    smtp_status(s,430,"problem executing spam control");
		    return;
		}
		written += wstat;
		towrite = strlen(msg)-written;
	    }
	    write(pipefd[1], msg, strlen(msg));

	    /* close the pipe */
	    close(pipefd[1]);

	    waitpid(checkpid, &status, 0);

	    if (WIFEXITED(status)) {
		/* normal exit of child process */
		int exitcode = WEXITSTATUS(status);
		switch (exitcode) {
		    case 42:
			log_error(ZONE, "could not execute spam check script");
			smtp_status(s,430,"problem executing spam control");
			return;
		    case 0: /* no spam mail */
			break;
		    case 1: /* the message is spam */
			log_debug(ZONE, "recevied message seems to be spam, rejecting");
			smtp_status(s,570,"this mail is probably spam and has not been accepted");
			return;
		    default: /* illegal return code */
			log_error(ZONE, "illegal exit code from spam check skript: %i", exitcode);
			smtp_status(s,430,"problem executing spam control");
			return;
		}
	    } else {
		log_error(ZONE, "spam check exited abnormally");
		smtp_status(s,430,"problem executing spam control");
		return;
	    }
	}
    } else {
	log_debug(ZONE, "not using spam check");
    }

    /* normalize line breaks to '\n' only */
    while ((ptr = strstr(msg, "\r\n")) != NULL) {
	strcpy(ptr, ptr+1);
    }
    while ((ptr = strchr(msg, '\r')) != NULL) {
	*ptr = '\n';
    }

    
    /* is this a message that uses MIME? */
    if (smtp_in_header(s, &ctx, msg, "MIME-Version", 1))
	ctx.mimeUsed = 1;
    ptr = smtp_in_header(s, &ctx, msg, "Content-Encoding", 1);
    if (!ptr)
	ctx.mimeEncoding = MIME_7BIT;
    else if (!strcasecmp(ptr, "8bit"))
	ctx.mimeEncoding = MIME_8BIT;
    else if (!strcasecmp(ptr, "7bit"))
	ctx.mimeEncoding = MIME_7BIT;
    else if (!strcasecmp(ptr, "quoted-printable"))
	ctx.mimeEncoding = MIME_QUOTED;
    else if (!strcasecmp(ptr, "base84"))
	ctx.mimeEncoding = MIME_BASE64;
    else if (!strcasecmp(ptr, "binary"))
	ctx.mimeEncoding = MIME_BINARY;

    x = xmlnode_new_tag("message");
    cur = jid_new(xmlnode_pool(x),s->i->i->id); /* jabber sending domain is based on configured service id */

    /* set the sending email addr frist from the reply-to, then the sender, then finally the addr smtp claims it was from */
    if((from = smtp_in_getaddr(s->p,smtp_in_header(s, &ctx, msg, "Reply-To", 1))) == NULL)
        if((from = smtp_in_getaddr(s->p,smtp_in_header(s, &ctx, msg, "From", 1))) == NULL)
            from = smtp_in_getaddr(s->p, s->from);
    jid_set(cur,from,JID_USER);
    xmlnode_put_attrib(x,"from",jid_full(cur));

    xmlnode_insert_cdata(xmlnode_insert_tag(x,"subject"),smtp_in_header(s,&ctx,  msg, "Subject", 0),-1);

    body = strstr(msg, "\n\n");

    /* XXX recode to UTF8 based on MIME hints */
    {
	char *ptr;
	for (ptr = body; *ptr; ptr++)
	    if (*ptr > 126 || *ptr < 1) *ptr = ' ';
    }
    
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"body"),body,-1);
    
    for(cur = s->rcpt; cur != NULL; cur = cur->next)
    {
        log_debug(ZONE,"sending to %s",jid_full(cur));
        xmlnode_put_attrib(x,"to",jid_full(cur));
        deliver(dpacket_new(xmlnode_dup(x)),s->i->i);
    }
    xmlnode_free(x);
    smtp_status(s,250,"Message Accepted");
}

/* read an incoming line from the connection */
void smtp_in_read(mio m, int state, void *arg, char *buffer,int bufsz)
{
    char *cbuffer, *cur, *args;
    smtps s = (smtps)(arg);
    int i;

    switch(state)
    {
        case MIO_NEW:
            s = smtp_new(m);
            s->phase = PHASE_CMD;
            s->host = mio_ip(m);
            s->i = (smtpi)arg; /* yeah, hack */
            log_debug(ZONE,"new connection from %s",s->host);
            smtp_status(s,220,"Jabber SMTP Transport");
            pool_cleanup(s->p, smtp_in_cleanup, (void *)s);
            mio_reset(m, smtp_in_read, (void *)s);
            return;
        case MIO_BUFFER:
            break;
        default:
            return;
    }

    s->timer = time(NULL); /* so we don't timeout */

    if(bufsz <= 0) return;

    /* XXX fix special characters, need more intelligence here */
    /*
    for(cur = buffer; (cur - buffer) < bufsz; cur++)
        if(*cur == 127 || (*cur < 32 && *cur != 10 && *cur != 13))
            *cur = '*';
    */

    for(i = 0; i < bufsz; i++)
       {
	   if (buffer[i] == '\0')
		  buffer[i]=' ';
	   }

    if(s->buffer != NULL)
    { /* old data yet, combine and use that */
        cur = malloc(strlen(s->buffer) + bufsz + 1);
        *cur = '\0';
        strcat(cur,s->buffer);
        strcat(cur,buffer);
        free(s->buffer);
        s->buffer = buffer = cur;
        bufsz = strlen(buffer);
    }

    log_debug(ZONE,"phase %d data [%s]",s->phase, buffer);

    cur = cbuffer = (char *)buffer;
    for(i = 1; i < bufsz; i++)
    {
        if(cbuffer[i-1] == '\r' && cbuffer[i] == '\n')
        {
            cbuffer[i-1] = '\0';
            cbuffer[i] = '\0';
            if(s->phase == PHASE_CMD)
            {
                if(strlen(cur) > 0)
                {
                    args = strstr(cur," ");
                    if(args != NULL)
                    {
                        *args = '\0';
                        ++args;
                    }
                    smtp_in_cmd(s, smtp_cmd_parse(cur), args);
                }
            }else if(s->phase == PHASE_DATA){
                if(*cur == '.')
                {
                    ++cur;
                    if(strlen(cur) == 0)
                    {
                        smtp_in_message(s);
                        s->x = NULL;
                        s->rcpt = NULL;
                        s->phase = PHASE_CMD;
                    }
                }
                if(s->phase == PHASE_DATA && cur)
                {
		    spooler(s->message, cur, "\r\n", s->message);
                }
            }
            cur = cbuffer + i + 1;
        }
    }

    cbuffer = NULL;
    i = strlen(cur);
    if(i > 0) /* the read was cut off */
    {
	if(i > 514)
	{
	    smtp_status(s,500,"XXX too much data");
	}else{
            cbuffer = strdup(cur);
        }
    }

    free(s->buffer);
    s->buffer = cbuffer; /* save the tail from above, if any */
}

