#include <time.h>
#include <stdio.h>
#include "smtp.h"

/* jabber->smtp */

char *_months[12]={"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

/* send an outgoing line to the connection */
void smtp_cmd(smtps s, int cmd, char *arg)
{
    char *out;
    pool p;
    spool line;

    p = pool_new();
    line = spool_new(p);

    switch(cmd)
    {
        case CMD_HELO:
            spool_add(line,"HELO ");
            break;

        case CMD_MAIL:
            spool_add(line,"MAIL FROM: ");
            break;

        case CMD_RCPT:
            spool_add(line,"RCPT TO: ");
            break;
        case CMD_DATA:
            spool_add(line,"DATA");
            break;
        case CMD_QUIT:
            spool_add(line,"QUIT");
            break;
    }
    if(arg != NULL)
    {
        if(cmd == CMD_HELO)
            spool_add(line,arg);
        else
            spooler(line,"<",arg,">",line);
    }

    spool_add(line,"\r\n");
    out = spool_print(line);
    s->lastcmd = cmd;
    log_debug("SMTP-SEND",out);
    mio_write(s->m,NULL,out,-1);
}

/* decode the incoming commands */
void smtp_in_status(smtps s, int status, char *data)
{
    char *lineend, *body;
    xmlnode x;

    log_debug(ZONE,"status %d response to command %d with data %s",status,s->lastcmd,data);

    switch(s->lastcmd)
    {
        case CMD_NONE:
            if(status != 220)
              break;
            smtp_cmd(s, CMD_HELO, s->i->me);
            return;

        case CMD_HELO:
            if(status != 250)
              break;
            smtp_cmd(s, CMD_MAIL, s->from);
            return;

        case CMD_MAIL:
            if(status != 250)
              break;
            smtp_cmd(s, CMD_RCPT, jid_full(s->rcpt));
            return;

        case CMD_RCPT:
            if(status != 250 && status != 251)
               break;
            smtp_cmd(s, CMD_DATA, NULL);
            return;

        case CMD_DATA:
            body = xmlnode_get_data(s->x);

            if (body == NULL)
               log_debug(ZONE,"Last Command DATA: body of message IS NULL");


            if(status != 354 || body == NULL)
              break;

            for(x = xmlnode_get_firstattrib(s->x); x != NULL; x = xmlnode_get_nextsibling(x))
               {
                mio_write(s->m,NULL,xmlnode_get_name(x),-1);
                mio_write(s->m,NULL,": ",2);
                mio_write(s->m,NULL,xmlnode_get_data(x),-1);
                mio_write(s->m,NULL,"\r\n",2);
               }
            mio_write(s->m,NULL,"\r\n",2);


	        while((lineend = strchr(body,'\n')) != NULL)
      	    {
//		     char *crpointer;

       		 lineend[0] = '\0';	/* cut the line */

             // S:2005-04-03 gorila@dione.zcu.cz
//		     while((crpointer = strchr(body, '\r')) != NULL)
//             crpointer[0] = ' ';
             // E:2005-04-03 gorila@dione.zcu.cz

		     if (body[0]=='.')
		        mio_write(s->m,NULL,".",1); /* escape the dot */

             // S:2005-04-03 gorila@dione.zcu.cz
             // xmlnode2str() return original XML data (without any changes)
             // xmlnode_get_tag_data() return data where CRLF and CR are replaced by LF
             // _mio_write_dump() hate data with length = 0
             // I added this exception
		     if (strlen(body)>0)
                mio_write(s->m,NULL,body,-1);
             // E:2005-04-03 gorila@dione.zcu.cz

		     mio_write(s->m,NULL,"\r\n",2);
		     body = lineend+1;	/* point to next line */
	        }

	        if (body[0])
		        mio_write(s->m,NULL,body,-1); /* print the rest */

            s->lastcmd = CMD_QUIT; /* so we know where we are */
            mio_write(s->m,NULL,"\r\n.\r\n",5);
            return;

        case CMD_QUIT:
            if(status != 250)
              break;
            smtp_cmd(s, CMD_QUIT, NULL);
            return;

         default:
             log_debug(ZONE,"SMTP relay connection not processed %d %d: %s",s->lastcmd,status,data);
    }
    log_debug(ZONE,"SMTP relay connection failed %d: %s",status,data);
}

/* when the connection dies, clean up the mess */
void smtp_out_cleanup(void *arg)
{
    smtps s = (smtps)(arg);
 
    log_debug(ZONE,"smtp_out_cleanup");
    if(s != NULL)
    {
        free(s->buffer);
    }
}

smtps smtp_out_new(mio m, jpacket p)
{
    smtpi i = (smtpi)xmlnode_get_vattrib(p->x,"smtp");
    smtps s;
    jid from;
    char *to, *host, *subject;
    time_t now_unixtime;
    struct tm now;
    char dateheader[27], smtpboundary[18], smtpCT[70];
    xmlnode html,meta;

    /* translate the jabber address to a sutiable SMTP sender */
    from = smtp_2mx(i, p->from);

    /* translate the recipient into a SMTP address */
    to = pstrdup(p->p,p->to->user);
    host = strstr(to,"%");
    *host = '@';
    ++host;

    log_debug(ZONE,"yay, sending a message from %s to %s [%s]",jid_full(from),to,xmlnode2str(p->x));
    s = smtp_new(m);
    s->phase = PHASE_SEND;
    s->host = mio_ip(m);
    s->i = i;
    s->rcpt = jid_new(s->p,to);
    s->from = pstrdup(s->p,jid_full(from));
    pool_cleanup(s->p, smtp_out_cleanup, (void *)s);

    /* create xml for outgoing message */
    s->x = xmlnode_new_tag_pool(s->p,"mail");

    xmlnode_put_attrib(s->x,"From",s->from);
    xmlnode_put_attrib(s->x,"To",to);

    subject = xmlnode_get_tag_data(p->x,"subject");

    xmlnode_put_attrib(s->x,"Subject",subject);
    xmlnode_put_attrib(s->x,"X-Mailer","Jabber-SMTP-Transport 0.5.3 [http://devel.amessage.de]");
    xmlnode_put_attrib(s->x,"MIME-Version","1.0");

    /* create time header for message */
    now_unixtime = time(NULL);
    gmtime_r(&now_unixtime, &now);
    dateheader[0]='\0';
    snprintf(dateheader, sizeof(dateheader), "%02i %s %04i %02i:%02i:%02i +0000", now.tm_mday, _months[now.tm_mon], now.tm_year+1900, now.tm_hour, now.tm_min, now.tm_sec);
    if (dateheader[0])
    {
	xmlnode_put_attrib(s->x,"Date",dateheader);
    }



    // xhtml and text
    if (xmlnode_get_tag(p->x,"html")!= NULL && xmlnode_get_tag(p->x,"body")!= NULL)
       {
       smtpboundary[0]='\0';
       snprintf(smtpboundary, sizeof(smtpboundary), "JPB_%04i%02i%02i%02i%02i%02i", now.tm_year+1900, now.tm_mon, now.tm_mday, now.tm_hour, now.tm_min, now.tm_sec);
       snprintf(smtpCT, sizeof(smtpCT), "multipart/alternative; boundary=\"%s\"", smtpboundary);

       xmlnode_put_attrib(s->x,"Content-type",smtpCT);

       xmlnode_insert_cdata(s->x,"--",-1);
       xmlnode_insert_cdata(s->x,smtpboundary,-1);
       xmlnode_insert_cdata(s->x,"\r\n",-1);
       xmlnode_insert_cdata(s->x,"Content-Type: text/plain; charset=UTF-8\r\n",-1);
       xmlnode_insert_cdata(s->x,"Content-Transfer-Encoding: 8bit\r\n",-1);
       xmlnode_insert_cdata(s->x,"Content-Description: Jabber message - text format\r\n",-1);
       xmlnode_insert_cdata(s->x,"\r\n",-1);
       xmlnode_insert_cdata(s->x,xmlnode_get_tag_data(p->x,"body"),-1);
       xmlnode_insert_cdata(s->x,"\r\n\r\n",-1);
       

       // for html is needly insert <head> tag with meta tag about charset
       // <head><meta content="text/html; charset=utf-8" http-equiv="content-type" /></head>
/*
       html=xmlnode_dup(xmlnode_get_tag(p->x,"html"));
       meta=xmlnode_insert_tag(xmlnode_insert_tag(html,"head"),"meta");
       xmlnode_put_attrib(meta,"content-type","text/html; charset=utf-8");
       xmlnode_put_attrib(meta,"http-equiv","content-type");
*/
       html= xmlnode_new_tag("html");
       if (xmlnode_has_attribs(xmlnode_get_tag(p->x,"html")))
           xmlnode_insert_node(html, xmlnode_get_firstattrib(xmlnode_get_tag(p->x,"html")));
       meta=xmlnode_insert_tag(xmlnode_insert_tag(html,"head"),"meta");
       xmlnode_put_attrib(meta,"content-type","text/html; charset=utf-8");
       xmlnode_put_attrib(meta,"http-equiv","content-type");
       xmlnode_insert_node(xmlnode_insert_tag(html,"body"), xmlnode_get_firstchild(xmlnode_get_tag(p->x,"html/body")));


       xmlnode_insert_cdata(s->x,"--",-1);
       xmlnode_insert_cdata(s->x,smtpboundary,-1);
       xmlnode_insert_cdata(s->x,"\r\n",-1);
       xmlnode_insert_cdata(s->x,"Content-Type: text/html; charset=UTF-8\r\n",-1);
       xmlnode_insert_cdata(s->x,"Content-Transfer-Encoding: 8bit\r\n",-1);
       xmlnode_insert_cdata(s->x,"Content-Description: Jabber message - html format\r\n",-1);
       xmlnode_insert_cdata(s->x,"\r\n",-1);
       xmlnode_insert_cdata(s->x,xmlnode2str(html),-1);
       xmlnode_insert_cdata(s->x,"\r\n\r\n",-1);
       
       xmlnode_insert_cdata(s->x,"--",-1);
       xmlnode_insert_cdata(s->x,smtpboundary,-1);
       xmlnode_insert_cdata(s->x,"--\r\n\r\n",-1);
       }


    // only xhtml
    if (xmlnode_get_tag(p->x,"html")!= NULL && xmlnode_get_tag(p->x,"body")== NULL)
       {
       html= xmlnode_new_tag("html");
       if (xmlnode_has_attribs(xmlnode_get_tag(p->x,"html")))
           xmlnode_insert_node(html, xmlnode_get_firstattrib(xmlnode_get_tag(p->x,"html")));
       meta=xmlnode_insert_tag(xmlnode_insert_tag(html,"head"),"meta");
       xmlnode_put_attrib(meta,"content-type","text/html; charset=utf-8");
       xmlnode_put_attrib(meta,"http-equiv","content-type");
       xmlnode_insert_node(xmlnode_insert_tag(html,"body"), xmlnode_get_firstchild(xmlnode_get_tag(p->x,"html/body")));

       xmlnode_put_attrib(s->x,"Content-Type","text/html; charset=UTF-8");
       xmlnode_put_attrib(s->x,"Content-Transfer-Encoding","8bit");
       xmlnode_put_attrib(s->x,"Content-Description","Jabber message - html format");
       xmlnode_insert_cdata(s->x,xmlnode2str(html),-1);
       }


    // only text
    if (xmlnode_get_tag(p->x,"html")== NULL && xmlnode_get_tag(p->x,"body")!= NULL)
       {
       xmlnode_put_attrib(s->x,"Content-Type","text/plain; charset=UTF-8");
       xmlnode_put_attrib(s->x,"Content-Transfer-Encoding","8bit");
       xmlnode_put_attrib(s->x,"Content-Description","Jabber message - text format");
       xmlnode_insert_cdata(s->x,xmlnode_get_tag_data(p->x,"body"),-1);
       }

    xmlnode_free(p->x);
    return s;
}

/* read an incoming line from the connection */
void smtp_out_read(mio m, int state, void *arg, char *buffer, int bufsz)
{
    char *cbuffer, *cur, *args;
    smtps s = (smtps)(arg);
    int i;

    switch(state)
    {
        case MIO_NEW:
            log_debug(ZONE,"new connection to relay");
            s = smtp_out_new(m, (jpacket)arg);
            mio_reset(m, smtp_out_read, (void *)s);
            return;
        case MIO_BUFFER:
            break;
        default:
            return;
    }

    s->timer = time(NULL); /* so we don't timeout */

    if(bufsz <= 0) return;

    if(s->buffer != NULL)
    { /* old data yet, combine and use that */
        cur = malloc(strlen(s->buffer) + bufsz + 1);
        *cur = '\0';
        strcat(cur,s->buffer);
        strcat(cur,buffer);
        free(s->buffer);
        s->buffer = buffer = cur;
        bufsz = strlen(buffer);
    }

    log_debug(ZONE,"phase %d data [%s]",s->phase, buffer);

    cur = cbuffer = (char *)buffer;
    for(i = 1; i < bufsz; i++)
    {
        if(cbuffer[i-1] == '\r' && cbuffer[i] == '\n')
        {
            cbuffer[i-1] = '\0';
            cbuffer[i] = '\0';
            if(strlen(cur) > 0)
            {
                args = strstr(cur," ");
                if(args != NULL)
                {
                    *args = '\0';
                    ++args;
                }
                smtp_in_status(s, atoi(cur), args);
            }
            cur = cbuffer + i + 1;
        }
    }

    free(s->buffer);
    s->buffer = NULL;
    i = strlen(cur);
    if(i > 0) /* the read was cut off */
    {
        s->buffer = strdup(cur);
    }
}


result smtp_out_packets(instance i, dpacket p, void *arg)
{
    smtpi si = (smtpi)arg;

    jid from;
    jpacket jp;

    log_debug(ZONE,"smtp getting a packet: %s",xmlnode2str(p->x));

    /* XXX check and make sure it's a message and stuff */
    jp = jpacket_new(p->x);
    from = smtp_2mx(si, jp->from);
    if(jp->to == NULL || jp->to->user == NULL || strstr(jp->to->user,"%") == NULL ||  from == NULL || jpacket_subtype(jp) == JPACKET__ERROR)
    {
        deliver_fail(p, "invalid address");
        return r_DONE;
    }

    /* hide session on the packet, and pass it to be handled after the conneciton is open */
    xmlnode_put_vattrib(jp->x, "smtp", arg);
    mio_connect(si->relay, 25, smtp_out_read, (void *)jp, 30, NULL, NULL);



    return r_DONE;
}
